from dataclasses import dataclass, field
from typing import Optional


@dataclass
class MembraneReset:
    """
    Reset the membrane potential v_mem to a given value
    after it spiked.

    Parameters:
        reset_value: fixed value that a neuron should be reset to. Defaults to zero.

    Example:
        >>> layer = sinabs.layers.LIF(reset_fn=MembraneReset(reset_value=0.), ...)
    """

    reset_value: float = 0.0

    def __call__(self, spikes, state, spike_threshold):
        new_state = state.copy()
        new_state["v_mem"] = (
            new_state["v_mem"] * (spikes == 0).float() + self.reset_value
        )
        return new_state


@dataclass
class MembraneSubtract:
    """
    Subtract the spiking threshold from the membrane potential
    for every neuron that spiked.

    Parameters:
        subtract_value: optional value that will be subtraced from
                        v_mem if it spiked. Defaults to spiking threshold if None.

    Example:
        >>> layer = sinabs.layers.LIF(reset_fn=MembraneSubtract(subtract_value=None), ...)
    """

    subtract_value: Optional[float] = None

    def __call__(self, spikes, state, threshold):
        new_state = state.copy()
        if self.subtract_value:
            new_state["v_mem"] = new_state["v_mem"] - spikes * self.subtract_value
        else:
            new_state["v_mem"] = new_state["v_mem"] - spikes * threshold
        return new_state



import torch
import torch.nn as nn
from sinabs.layers import StatefulLayer, LIF
import pytest
import numpy as np


def test_stateful_layer():
    layer = StatefulLayer(state_names=["v_mem"])
    assert not layer.is_state_initialised()
    assert layer.v_mem.shape == torch.zeros((0)).shape

    layer.reset_states()
    assert not layer.is_state_initialised()
    assert layer.v_mem.shape == torch.zeros((0)).shape


def test_lif_reset():
    layer = LIF(tau_mem=30.0)
    assert not layer.is_state_initialised()
    assert layer.v_mem.shape == torch.zeros((0)).shape

    layer(torch.rand((2, 3, 4)))
    assert layer.is_state_initialised()
    assert layer.v_mem.shape == (2, 4)
    assert layer.v_mem.sum() != 0

    layer.reset_states()
    assert layer.is_state_initialised()
    assert layer.v_mem.shape == (2, 4)
    assert layer.v_mem.sum() == 0

    layer.reset_states(randomize=True)
    assert layer.is_state_initialised()
    assert layer.v_mem.shape == (2, 4)
    assert layer.v_mem.sum() != 0


def test_changing_batch_size():
    layer = LIF(tau_mem=30.0)
    assert not layer.is_state_initialised()
    assert layer.v_mem.shape == torch.zeros((0)).shape

    layer(torch.rand((2, 3, 4)))
    assert layer.is_state_initialised()
    assert layer.v_mem.shape == (2, 4)

    layer(torch.rand((5, 3, 4)))
    assert layer.is_state_initialised()
    assert layer.v_mem.shape == (5, 4)


def test_reset_states():
    layer = StatefulLayer(state_names=["v_mem", "i_syn"])
    layer.init_state_with_shape((1, 2, 3))
    assert (1, 2, 3) == layer.v_mem.shape
    assert (1, 2, 3) == layer.i_syn.shape
    # Reset states
    layer.reset_states(randomize=False)
    assert layer.v_mem.any() == False
    # Reset states
    layer.reset_states(randomize=True)
    assert layer.v_mem.any() == True

    # Reset states
    layer.reset_states(randomize=True, value_ranges={"v_mem": (-5, -3)})
    assert layer.v_mem.max() <= -3
    assert layer.v_mem.min() >= -5

    assert layer.i_syn.max() <= 1.0
    assert layer.i_syn.min() >= 0.0



from typing import Union, List, Tuple
import numpy as np
from torch import nn

ArrayLike = Union[np.ndarray, List, Tuple]


class Cropping2dLayer(nn.Module):
    """
    Crop input image by
    """

    def __init__(
        self,
        cropping: ArrayLike = ((0, 0), (0, 0)),
    ):
        """
        Crop input to the the rectangle dimensions

        :param cropping: ((top, bottom), (left, right))
        """
        super().__init__()
        self.top_crop, self.bottom_crop = cropping[0]
        self.left_crop, self.right_crop = cropping[1]

    def forward(self, binary_input):
        _, self.channels_in, h, w = list(binary_input.shape)
        # Crop the data array
        crop_out = binary_input[
            :,
            :,
            self.top_crop : h - self.bottom_crop,
            self.left_crop : w - self.right_crop,
        ]
        self.out_shape = crop_out.shape[1:]
        self.spikes_number = crop_out.abs().sum()
        self.tw = len(crop_out)
        return crop_out

    def get_output_shape(self, input_shape: Tuple) -> Tuple:
        """
        Retuns the output dimensions

        :param input_shape: (channels, height, width)
        :return: (channels, height, width)
        """
        channels, height, width = input_shape
        return (
            channels,
            height - self.top_crop - self.bottom_crop,
            width - self.left_crop - self.right_crop,
        )



"""
===========
MaxSpike
===========
:class:`~sinabs.activation.MaxSpike` activation function.
"""

import torch
import sinabs.activation as sina
import matplotlib.pyplot as plt

v_mem = torch.linspace(0, 5.5, 500)

spike_threshold = 1.0
activations = sina.MaxSpike(max_num_spikes_per_bin=2).apply(
    v_mem, spike_threshold, sina.MultiGaussian()
)
plt.plot(v_mem, activations)
plt.xlabel("Neuron membrane potential")
plt.ylabel("Spike activation")
plt.ylim(top=5.2)
plt.tight_layout()



import onnx
import torch
from sinabs import Network


def print_graph(graph: onnx.GraphProto):
    params = [x.name for x in graph.initializer]
    for i, node in enumerate(graph.node):
        inputs = list(filter(lambda x: x not in params, node.input))
        print(f"{node.name}: Inputs({inputs}), Outputs({node.output}")


def print_onnx_model(onnx_model: onnx.ModelProto):
    print_graph(onnx_model.graph)


def get_graph(model: Network, inputs):
    """
    Extract graph from a sinabs Network model

    :param model: sinabs.Netowrk model to extract the graph for
    :param inputs: Input tensor to extract graph
    :return: onnx.Graph
    """

    fname = "ann_temp.onnx"
    if len(inputs.shape) == 5:
        inputs = inputs.sum(0)
    torch.onnx.export(model.analog_model, inputs, fname)
    onnx_model = onnx.load(fname)
    return onnx_model.graph



import torch.nn as nn
from sinabs.activation import Quantize


class QuantizeLayer(nn.Module):
    """
    Layer that quantizes the input, i.e. returns floor(input).

    :param quantize: If False, this layer will do nothing.
    """

    def __init__(self, quantize=True):
        super().__init__()
        self.quantize = quantize

    def forward(self, data):
        if self.quantize:
            return Quantize.apply(data)
        else:
            return data



import torch
import torch.nn as nn
import numpy as np
from typing import List


def get_activations(torchanalog_model, tsrData, name_list=None):
    """
    Return torch analog model activations for the specified layers
    """
    torch_modules = dict(torchanalog_model.named_modules())

    # Populate layer names
    if name_list is None:
        name_list = ["Input"] + list(torch_modules.keys())[1:]

    analog_activations = []

    for layer_name in name_list:
        if layer_name == "Input":
            # Bypass input layers
            arrOut = tsrData.detach().cpu().numpy()
            analog_activations.append(arrOut)

    # Define hook
    def hook(module, inp, output):
        arrOut = output.detach().cpu().numpy()
        analog_activations.append(arrOut)

    hooklist = []
    # Attach and register hook
    for layer_name in name_list:
        if layer_name == "Input":
            continue
        torch_layer = torch_modules[layer_name]
        hooklist.append(torch_layer.register_forward_hook(hook))

    # Do a forward pass
    with torch.no_grad():
        torchanalog_model.eval()
        torchanalog_model(tsrData)

    # Remove hooks
    for h in hooklist:
        h.remove()

    return analog_activations


def get_network_activations(
    model: nn.Module, inp, name_list: List = None, bRate: bool = False
) -> [np.ndarray]:
    """
    Returns the activity of neurons in each layer of the network

    :param model: Model for which the activations are to be read out
    :param inp: Input to the model
    :param bRate: If true returns the rate, else returns spike count
    :param name_list: list of all layers whose activations need to be compared
    """
    spike_counts = []
    tSim = len(inp)

    # Define hook
    def hook(module, inp, output):
        arrOut = output.float().sum(0).cpu().numpy()
        spike_counts.append(arrOut)

    # Generate default list of layers
    if name_list is None:
        name_list = ["Input"] + [lyr.layer_name for lyr in model.layers]

    # Extract activity for each layer of interest
    for layer_name in name_list:
        # Append input activity
        if layer_name == "Input":
            spike_counts.append(inp.float().sum(0).cpu().numpy() * 1000)
        else:
            # Activity of other layers
            lyr = dict(model.named_modules())[layer_name]
            lyr.register_forward_hook(hook)

    with torch.no_grad():
        model(inp)

    if bRate:
        spike_counts = [(counts / tSim * 1000) for counts in spike_counts]
    return spike_counts


def normalize_weights(
    ann: nn.Module,
    sample_data: torch.Tensor,
    output_layers: List[str],
    param_layers: List[str],
    percentile: float = 99,
):
    """
    Rescale the weights of the network, such that the activity of each specified layer is normalized.

    The method implemented here roughly follows the paper:
    `Conversion of Continuous-Valued Deep Networks to Efficient Event-Driven Networks for Image Classification` by Rueckauer et al.
    https://www.frontiersin.org/article/10.3389/fnins.2017.00682

    Args:
         ann(nn.Module): Torch module
         sample_data (nn.Tensor): Input data to normalize the network with
         output_layers (List[str]): List of layers to verify activity of normalization. Typically this is a relu layer
         param_layers (List[str]): List of layers whose parameters preceed `output_layers`
         percentile (float): A number between 0 and 100 to determine activity to be normalized by.
          where a 100 corresponds to the max activity of the network. Defaults to 99.
    """
    # Network activity storage
    output_data = []

    # Hook to save data
    def save_data(lyr, input, output):
        output_data.append(output.clone())

    # All the named layers of the module
    named_layers = dict(ann.named_children())

    for i in range(len(output_layers)):
        param_layer = named_layers[param_layers[i]]
        output_layer = named_layers[output_layers[i]]

        handle = output_layer.register_forward_hook(save_data)

        with torch.no_grad():
            _ = ann(sample_data)

            # Get max output
            max_lyr_out = np.percentile(output_data[-1].cpu().numpy(), percentile)

            # Rescale weights to normalize max output
            for p in param_layer.parameters():
                p.data *= 1 / max_lyr_out

        output_data.clear()
        # Deregister hook
        handle.remove()



import torch
import torch.nn as nn
from typing import Optional, Union, Callable
from sinabs.activation import MultiSpike, MembraneSubtract, SingleExponential
from . import functional
from .stateful_layer import StatefulLayer
from .reshape import SqueezeMixin


class LIF(StatefulLayer):
    """
    Leaky Integrate and Fire neuron layer.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = \\alpha V_{mem}(t) + (1-\\alpha)\\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\alpha =  e^{-1/tau_{mem}}` and :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters
    ----------
    tau_mem: float
        Membrane potential time constant.
    tau_syn: float
        Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
    spike_threshold: float
        Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
    spike_fn: torch.autograd.Function
        Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
        a spike threshold and a surrogate gradient function and returns spikes. Be aware
        that the class itself is passed here (because torch.autograd methods are static)
        rather than an object instance.
    reset_fn: Callable
        A function that defines how the membrane potential is reset after a spike.
    surrogate_grad_fn: Callable
        Choose how to define gradients for the spiking non-linearity during the
        backward pass. This is a function of membrane potential.
    min_v_mem: float or None
        Lower bound for membrane potential v_mem, clipped at every time step.
    train_alphas: bool
        When True, the discrete decay factor exp(-1/tau) is used for training rather than tau itself.
    shape: torch.Size
        Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
    norm_input: bool
        When True, normalise input current by tau. This helps when training time constants.
    record_states: bool
        When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.
    """

    def __init__(
        self,
        tau_mem: Union[float, torch.Tensor],
        tau_syn: Optional[Union[float, torch.Tensor]] = None,
        spike_threshold: float = 1.0,
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        min_v_mem: Optional[float] = None,
        train_alphas: bool = False,
        shape: Optional[torch.Size] = None,
        norm_input: bool = True,
        record_states: bool = False,
    ):
        super().__init__(
            state_names=["v_mem", "i_syn"] if tau_syn is not None else ["v_mem"]
        )
        if train_alphas:
            self.alpha_mem = nn.Parameter(
                torch.exp(-1.0 / torch.as_tensor(tau_mem, dtype=torch.float32))
            )
            self.alpha_syn = (
                nn.Parameter(
                    torch.exp(-1.0 / torch.as_tensor(tau_syn, dtype=torch.float32))
                )
                if tau_syn is not None
                else None
            )
        else:
            self.tau_mem = nn.Parameter(torch.as_tensor(tau_mem, dtype=torch.float32))
            self.tau_syn = (
                nn.Parameter(torch.as_tensor(tau_syn, dtype=torch.float32))
                if tau_syn is not None
                else None
            )
        self.spike_threshold = spike_threshold
        self.spike_fn = spike_fn
        self.reset_fn = reset_fn
        self.surrogate_grad_fn = surrogate_grad_fn
        self.min_v_mem = min_v_mem
        self.train_alphas = train_alphas
        self.norm_input = norm_input
        self.record_states = record_states
        if shape:
            self.init_state_with_shape(shape)

    @property
    def alpha_mem_calculated(self):
        if self.train_alphas:
            return self.alpha_mem
        else:
            return torch.exp(-1.0 / self.tau_mem)

    @property
    def alpha_syn_calculated(self):
        if self.train_alphas:
            return self.alpha_syn
        elif self.tau_syn is not None:
            # Calculate alpha with 64 bit floating point precision
            return torch.exp(-1.0 / self.tau_syn)
        else:
            return None

    @property
    def tau_mem_calculated(self):
        if self.train_alphas:
            if self.alpha_mem is None:
                return None
            else:
                return -1.0 / torch.log(self.alpha_mem)
        else:
            return self.tau_mem

    @property
    def tau_syn_calculated(self):
        if self.train_alphas:
            if self.alpha_syn is None:
                return None
            else:
                return -1.0 / torch.log(self.alpha_syn)
        else:
            return self.tau_syn

    def forward(self, input_data: torch.Tensor):
        """
        Forward pass with given data.

        Parameters:
            input_current : torch.Tensor
                Data to be processed. Expected shape: (batch, time, ...)

        Returns:
            torch.Tensor
                Output data. Same shape as `input_data`.
        """
        batch_size, time_steps, *trailing_dim = input_data.shape

        # Ensure the neuron state are initialized
        if not self.is_state_initialised() or not self.state_has_shape(
            (batch_size, *trailing_dim)
        ):
            self.init_state_with_shape((batch_size, *trailing_dim))

        alpha_mem = self.alpha_mem_calculated
        alpha_syn = self.alpha_syn_calculated

        spikes, state, recordings = functional.lif_forward(
            input_data=input_data,
            alpha_mem=alpha_mem,
            alpha_syn=alpha_syn,
            state=dict(self.named_buffers()),
            spike_threshold=self.spike_threshold,
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            min_v_mem=self.min_v_mem,
            norm_input=self.norm_input,
            record_states=self.record_states,
        )
        self.v_mem = state["v_mem"]
        self.i_syn = state["i_syn"] if alpha_syn is not None else None
        self.recordings = recordings

        self.firing_rate = spikes.sum() / spikes.numel()
        return spikes

    @property
    def shape(self):
        if self.is_state_initialised():
            return self.v_mem.shape
        else:
            return None

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        tau_syn = self.tau_syn_calculated
        if tau_syn is not None:
            tau_syn = tau_syn.detach()
        tau_mem = self.tau_mem_calculated
        if tau_mem is not None:
            tau_mem = tau_mem.detach()

        param_dict.update(
            tau_mem=tau_mem,
            tau_syn=tau_syn,
            spike_threshold=self.spike_threshold,
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            train_alphas=self.train_alphas,
            shape=self.shape,
            min_v_mem=self.min_v_mem,
            norm_input=self.norm_input,
            record_states=self.record_states,
        )
        return param_dict


class LIFRecurrent(LIF):
    """
    Leaky Integrate and Fire neuron layer with recurrent connections.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = \\alpha V_{mem}(t) + (1-\\alpha)\\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\alpha =  e^{-1/tau_{mem}}` and :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters
    ----------
    tau_mem: float
        Membrane potential time constant.
    rec_connect: torch.nn.Module
        An nn.Module which defines the recurrent connectivity, e.g. nn.Linear
    tau_syn: float
        Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
    spike_threshold: float
        Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
    activation_fn: Callable
        a torch.autograd.Function to provide forward and backward calls. Takes care of all the spiking behaviour.
    min_v_mem: float or None
        Lower bound for membrane potential v_mem, clipped at every time step.
    train_alphas: bool
        When True, the discrete decay factor exp(-1/tau) is used for training rather than tau itself.
    shape: torch.Size
        Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
    norm_input: bool
        When True, normalise input current by tau. This helps when training time constants.
    record_states: bool
        When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.
    """

    def __init__(
        self,
        tau_mem: Union[float, torch.Tensor],
        rec_connect: torch.nn.Module,
        tau_syn: Optional[Union[float, torch.Tensor]] = None,
        spike_threshold: float = 1.0,
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        min_v_mem: Optional[float] = None,
        train_alphas: bool = False,
        shape: Optional[torch.Size] = None,
        norm_input: bool = True,
        record_states: bool = False,
    ):
        super().__init__(
            tau_mem=tau_mem,
            tau_syn=tau_syn,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            train_alphas=train_alphas,
            norm_input=norm_input,
            record_states=record_states,
        )
        self.rec_connect = rec_connect

    def forward(self, input_data: torch.Tensor):
        """
        Forward pass with given data.

        Parameters:
            input_current : torch.Tensor
                Data to be processed. Expected shape: (batch, time, ...)

        Returns:
            torch.Tensor
                Output data. Same shape as `input_data`.
        """
        batch_size, time_steps, *trailing_dim = input_data.shape

        # Ensure the neuron state are initialized
        if not self.is_state_initialised() or not self.state_has_shape(
            (batch_size, *trailing_dim)
        ):
            self.init_state_with_shape((batch_size, *trailing_dim))

        alpha_mem = self.alpha_mem_calculated
        alpha_syn = self.alpha_syn_calculated

        spikes, state, recordings = functional.lif_recurrent(
            input_data=input_data,
            alpha_mem=alpha_mem,
            alpha_syn=alpha_syn,
            state=dict(self.named_buffers()),
            spike_threshold=self.spike_threshold,
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            min_v_mem=self.min_v_mem,
            norm_input=self.norm_input,
            rec_connect=self.rec_connect,
            record_states=self.record_states,
        )
        self.v_mem = state["v_mem"]
        self.i_syn = state["i_syn"] if alpha_syn is not None else None
        self.recordings = recordings

        self.firing_rate = spikes.sum() / spikes.numel()
        return spikes

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.update(rec_connect=self.rec_connect)
        return param_dict


class LIFSqueeze(LIF, SqueezeMixin):
    """
    LIF layer with 4-dimensional input (Batch*Time, Channel, Height, Width).

    Same as parent LIF class, only takes in squeezed 4D input (Batch*Time, Channel, Height, Width)
    instead of 5D input (Batch, Time, Channel, Height, Width) in order to be compatible with
    layers that can only take a 4D input, such as convolutional and pooling layers.
    """

    def __init__(
        self,
        batch_size=None,
        num_timesteps=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.squeeze_init(batch_size, num_timesteps)

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        return self.squeeze_forward(input_data, super().forward)

    @property
    def _param_dict(self) -> dict:
        return self.squeeze_param_dict(super()._param_dict)



import warnings

import torch
import torch.nn as nn
import numpy as np
from typing import Optional, Union, List, Tuple, Dict
from .utils import get_network_activations, get_activations
from .layers import StatefulLayer
from .synopcounter import SNNSynOpCounter

ArrayLike = Union[np.ndarray, List, Tuple]


class Network(torch.nn.Module):
    """
    Class of a spiking neural network

    Attributes:
        spiking_model: torch.nn.Module, a spiking neural network model
        analog_model: torch.nn.Module, an artifical neural network model
        input_shape: Tuple, size of input
        synops: If True (default: False), register hooks for counting synaptic \
        operations during forward passes, instantiating `sinabs.SNNSynOpCounter`.
    """

    def __init__(
        self,
        analog_model=None,
        spiking_model=None,
        input_shape: Optional[ArrayLike] = None,
        synops: bool = False,
        batch_size: int = 1,
        num_timesteps: int = 1,
    ):
        super().__init__()
        self.spiking_model: nn.Module = spiking_model
        self.analog_model: nn.Module = analog_model
        self.input_shape = input_shape

        self.synops = synops
        if synops:
            self.synops_counter = SNNSynOpCounter(self.spiking_model)

        if input_shape is not None and spiking_model is not None:
            self._compute_shapes(
                input_shape, batch_size=batch_size, num_timesteps=num_timesteps
            )

    @property
    def layers(self):
        return list(self.spiking_model.named_children())

    def _compute_shapes(self, input_shape, batch_size=1, num_timesteps=1):
        def hook(module, inp, out):
            module.out_shape = out.shape[1:]

        hook_list = []
        for layer in self.spiking_model.modules():
            this_hook = layer.register_forward_hook(hook)
            hook_list.append(this_hook)

        device = next(self.parameters()).device

        # Infer shape
        if batch_size is None:
            batch_size = 1
        if num_timesteps is None:
            num_timesteps = 1
        shape = [batch_size * num_timesteps] + list(input_shape)
        dummy_input = torch.zeros(shape, requires_grad=False).to(device)
        # do a forward pass
        self(dummy_input)

        [this_hook.remove() for this_hook in hook_list]

    def forward(self, tsrInput) -> torch.Tensor:
        """
        Forward pass for this model
        """
        return self.spiking_model(tsrInput)

    def compare_activations(
        self,
        data,
        name_list: Optional[ArrayLike] = None,
        compute_rate: bool = False,
        verbose: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray, str]:
        """
        Compare activations of the analog model and the SNN for a given data sample

        Args:
            data (np.ndarray):      Data to process
            name_list (List[str]):  list of all layer names (str) whose activations need to be compared
            compute_rate (bool):    True if you want to compute firing rate. By default spike count is returned
            verbose (bool):         print debugging logs to the terminal
        Returns:
            tuple: A tuple of lists (ann_activity, snn_activity, name_list)
                - ann_activity: output activity of the ann layers
                - snn_activity: output activity of the snn layers
                - name_list: spiking layers' name list for plotting comparison
        """
        if name_list is None:
            name_list = ["Input"]
            for layer_name, lyr in self.spiking_model.named_modules():
                if isinstance(lyr, StatefulLayer):
                    name_list.append(layer_name)

        if verbose:
            print("Comparing activations for {0}".format(name_list))

        # Calculate activations for the torch analog model
        if compute_rate:
            tsrAnalogData = data.mean(0).unsqueeze(0)
        else:
            tsrAnalogData = data.sum(0).unsqueeze(0)
        with torch.no_grad():
            analog_activations = get_activations(
                self.analog_model, tsrAnalogData, name_list=name_list
            )

        # Calculate activations for spiking model
        spike_rates = get_network_activations(
            self.spiking_model, data, name_list=name_list, bRate=compute_rate
        )
        return analog_activations, spike_rates, name_list

    def plot_comparison(
        self, data, name_list: Optional[ArrayLike] = None, compute_rate=False
    ):
        """
        Plots a scatter plot of all the activations

        Args:
            data: Data to be processed
            name_list: ArrayLike with names of all the layers of interest to be compared
            compute_rate: Compare firing rates instead of spike count
        Returns:
            tuple: A tuple of lists (ann_activity, snn_activity)
                - ann_activity: output activity of the ann layers
                - snn_activity: output activity of the snn layers
        """
        import pylab

        analog_activations, spike_rates, name_list = self.compare_activations(
            data, name_list=name_list, compute_rate=compute_rate
        )
        for nLyrIdx in range(len(name_list)):
            pylab.scatter(
                spike_rates[nLyrIdx],
                analog_activations[nLyrIdx],
                label=name_list[nLyrIdx],
            )
        if compute_rate:
            pylab.xlabel("Spike rates (Hz)")
        else:
            pylab.xlabel("# Spike count")
        pylab.ylabel("Analog activations")
        pylab.legend()
        return analog_activations, spike_rates

    def reset_states(
        self,
        randomize: bool = False,
        value_ranges: Optional[List[Dict[str, Tuple[float, float]]]] = None,
    ):
        """
        Reset all neuron states in the submodules.

        Parameters
        ----------
        randomize: Bool
            If true, reset the states between a range provided. Else, the states are reset to zero.
        value_ranges: Optional[List[Dict[str, Tuple[float, float]]]]
            A list of value_range dictionaries with the same length as the total stateful layers in the module.
            Each dictionary is a key value pair: buffer_name -> (min, max) for each state that needs to be reset.
            The states are reset with a uniform distribution between the min and max values specified.
            Any state with an undefined key in this dictionary will be reset between 0 and 1
            This parameter is only used if randomize is set to true.
        """

        if value_ranges:
            num_stateful_layers = len(
                [None for mod in self.modules() if isinstance(mod, StatefulLayer)]
            )
            if len(value_ranges) != num_stateful_layers:
                raise TypeError(
                    "The number of entries in value_ranges does not match the number of stateful sub modules"
                )
        i = 0
        for lyr in self.modules():
            if isinstance(lyr, StatefulLayer):
                if value_ranges is None:
                    vr = None
                else:
                    vr = value_ranges[i]
                    i += 1
                lyr.reset_states(randomize=randomize, value_ranges=vr)

    def get_synops(self, num_evs_in=None) -> dict:
        """
        Please see docs for `sinabs.SNNSynOpCounter.get_synops()`.
        """
        if num_evs_in is not None:
            warnings.warn("num_evs_in is deprecated and has no effect")

        return self.synops_counter.get_synops()


def get_parent_module_by_name(
    root: torch.nn.Module, name: str
) -> Tuple[torch.nn.Module, str]:
    """
    Find a nested Module of a given name inside a Module, and return its parent
    Module.

    Args:
        root: The Module inside which to look for the nested Module
        name: Name of the Module that is being searched for within root. Must
              contain all parent modules, separated by a `.` , e.g.
              "root.nested_module1.nested_module2.desired_module"
    Returns:
        torch.nn.Module: The Module that contains the Module with the given name. In the example
        above this would be `nested_module2`.
        str: The name of the child, without parent modules, e.g. "desired_module"
    """
    if "." not in name:
        if not hasattr(root, name):
            raise KeyError(f"The requested module `{name}` could not be found.")

        return root, name
    else:
        child_name, *rest = name.split(".")
        child = getattr(root, child_name)
        return get_parent_module_by_name(child, ".".join(rest))


def infer_module_device(module: torch.nn.Module) -> Union[torch.device, None]:
    """
    Infere on which device a module is operating by first looking at its parameters
    and then, if no parameters are found, at its buffers.

    Args:
        module: The module whose device is to be inferred.

    Returns:
        torch.device: The device of 'module', or `None` if no device has been found.
    """

    try:
        return next(module.parameters()).device
    except StopIteration:
        # No parameters, try buffers
        try:
            return next(module.buffers()).device
        except StopIteration:
            # No buffers, don't infer device
            return None



from sinabs import SynOpCounter, SNNSynOpCounter
from sinabs.layers import NeuromorphicReLU, IAF
import torch
import numpy as np
from torch import nn


class Model(torch.nn.Module):
    def __init__(self, quantize=False):
        super().__init__()

        self.seq = torch.nn.Sequential(
            torch.nn.Conv2d(
                in_channels=1, out_channels=8, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Conv2d(
                in_channels=8, out_channels=12, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Conv2d(
                in_channels=12, out_channels=12, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Dropout2d(0.5),
            torch.nn.Flatten(),
            torch.nn.Linear(432, 10, bias=False),
            NeuromorphicReLU(fanout=0),
        )

    def forward(self, x):
        return self.seq(x)


class TinyModel(torch.nn.Module):
    def __init__(self, quantize):
        super().__init__()
        self.linear = torch.nn.Linear(5, 2, bias=False)
        self.relu = NeuromorphicReLU(fanout=2, quantize=quantize)

        self.linear.weight.data = torch.tensor(
            [[1.2, 1.0, 1.0, 3.0, -2.0], [1.2, 1.0, 1.0, 2.0, -10.0]]
        )

    def forward(self, x):
        return self.relu(self.linear(x))


def test_parsing():
    model = Model()
    loss = SynOpCounter(model.modules())

    assert len(loss.modules) == 3


def test_loss():
    model = TinyModel(quantize=False)
    criterion = SynOpCounter(model.modules())

    input = torch.tensor([[0.5, 0.5, 0.5, 0.5, 0.5]])
    model(input)

    loss = criterion()

    assert np.allclose(loss.item(), 4.2)


def test_loss_quantized():
    model = TinyModel(quantize=True)
    criterion = SynOpCounter(model.modules())

    input = torch.tensor([[0.5, 0.5, 0.5, 0.5, 0.5]])
    model(input)

    loss = criterion()

    assert np.allclose(loss.item(), 4.0)


def test_layer_synops():
    model = Model()
    criterion = SynOpCounter(model.modules(), sum_activations=False)

    input = torch.rand([1, 1, 64, 64])
    model(input)

    loss = criterion()

    assert len(loss) == 3


def test_snn_synops_counter():
    model = nn.Sequential(nn.Conv2d(1, 5, kernel_size=2), IAF())

    inp = torch.tensor([[[[0, 0, 0], [0, 3, 0], [0, 0, 0]]]]).float()

    counter = SNNSynOpCounter(model)
    model(inp)
    # 3 spikes, 2x2 kernel, 5 channels
    assert sum(x["SynOps"] for x in counter.get_synops().values()) == 60
    assert counter.get_total_synops() == 60



def test_maxpool2d():
    from sinabs.layers import SpikingMaxPooling2dLayer
    import torch

    inp = torch.zeros((2, 2, 10, 10))
    maxpool = SpikingMaxPooling2dLayer(pool_size=(2, 2), strides=(2, 2))

    # Verify output shape
    assert maxpool(inp).shape == (2, 2, 5, 5)


def test_maxpool_function():
    from sinabs.layers import SpikingMaxPooling2dLayer
    import torch

    lyr = SpikingMaxPooling2dLayer(pool_size=(2, 3), strides=None)

    tsrInput = (torch.rand(10, 1, 2, 3) > 0.8).float()
    # print(tsrInput.sum(0))
    tsrInput[:, 0, 0, 2] = 1
    tsrOut = lyr(tsrInput)

    assert tsrOut.sum().item() == 10.0



"""
=============
MultiGaussian
=============
:class:`~sinabs.activation.MultiGaussian` surrogate gradient.
"""

import torch
import sinabs.activation as sina
import matplotlib.pyplot as plt

x = torch.linspace(-0.5, 3.5, 500)
plt.plot(x, sina.MultiGaussian()(v_mem=x, spike_threshold=1.0), label="MultiGaussian")
plt.xlabel("Neuron membrane potential")
plt.ylabel("Derivative")
plt.legend()



def test_img2spk():
    import torch
    from sinabs.layers import Img2SpikeLayer

    lyr = Img2SpikeLayer(
        image_shape=(2, 64, 64),
        tw=10,
        max_rate=1000,
    )

    img = torch.rand(2, 64, 64)

    spks = lyr(img)

    assert spks.shape == (10, 2, 64, 64)



import torch
import pytest
from torch import nn
from sinabs.utils import normalize_weights


# init model
CNN = nn.Sequential(
    nn.Conv2d(1, 16, kernel_size=(3, 3), bias=False),
    nn.BatchNorm2d(16, affine=True),
    nn.ReLU(),
    nn.AvgPool2d(kernel_size=2),
    nn.Conv2d(16, 8, kernel_size=(3, 3), bias=True),
    nn.BatchNorm2d(8, affine=False, track_running_stats=True),
)

# get activation layers and parameter layers
SPIKE_LAYERS = [
    name for name, child in CNN.named_children() if isinstance(child, (nn.ReLU,))
]
PARAM_LAYERS = [
    name
    for name, child in CNN.named_children()
    if isinstance(child, (nn.Conv2d, nn.Linear))
]


def test_normalize_weights_cpu():
    # test on cpu tensor
    sample_input_cpu = torch.rand(8, 1, 28, 28)
    normalize_weights(
        CNN.cpu(), sample_input_cpu, SPIKE_LAYERS, SPIKE_LAYERS, percentile=100
    )


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_normalize_weights_gpu():
    # test on cuda tensor
    sample_input_gpu = torch.rand(8, 1, 28, 28).cuda()
    normalize_weights(
        CNN.cuda(), sample_input_gpu, SPIKE_LAYERS, SPIKE_LAYERS, percentile=100
    )



import torch
from copy import deepcopy
from typing import Optional, Callable
from sinabs.activation import MultiSpike, MembraneSubtract, SingleExponential
from .reshape import SqueezeMixin
from .lif import LIF, LIFRecurrent
import numpy as np


class IAF(LIF):
    """
    Integrate and Fire neuron layer.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = V_{mem}(t) + \\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters
    ----------
    spike_threshold: float
        Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
    spike_fn: torch.autograd.Function
        Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
        a spike threshold and a surrogate gradient function and returns spikes. Be aware
        that the class itself is passed here (because torch.autograd methods are static)
        rather than an object instance.
    reset_fn: Callable
        A function that defines how the membrane potential is reset after a spike.
    surrogate_grad_fn: Callable
        Choose how to define gradients for the spiking non-linearity during the
        backward pass. This is a function of membrane potential.
    tau_syn: float
        Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
    min_v_mem: float or None
        Lower bound for membrane potential v_mem, clipped at every time step.
    shape: torch.Size
        Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
    record_states: bool
        When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.
    """

    def __init__(
        self,
        spike_threshold: float = 1.0,
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        tau_syn: Optional[float] = None,
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        record_states: bool = False,
    ):
        super().__init__(
            tau_mem=np.inf,
            tau_syn=tau_syn,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            norm_input=False,
            record_states=record_states,
        )
        # IAF does not have time constants
        self.tau_mem = None

    @property
    def alpha_mem_calculated(self):
        return torch.tensor(1.0)

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.pop("tau_mem")
        param_dict.pop("train_alphas")
        param_dict.pop("norm_input")
        return param_dict


class IAFRecurrent(LIFRecurrent):
    """
    Integrate and Fire neuron layer with recurrent connections.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = V_{mem}(t) + \\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters
    ----------
    rec_connect: torch.nn.Module
        An nn.Module which defines the recurrent connectivity, e.g. nn.Linear
    spike_threshold: float
        Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
    spike_fn: torch.autograd.Function
        Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
        a spike threshold and a surrogate gradient function and returns spikes. Be aware
        that the class itself is passed here (because torch.autograd methods are static)
        rather than an object instance.
    reset_fn: Callable
        A function that defines how the membrane potential is reset after a spike.
    surrogate_grad_fn: Callable
        Choose how to define gradients for the spiking non-linearity during the
        backward pass. This is a function of membrane potential.
    tau_syn: float
        Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
    min_v_mem: float or None
        Lower bound for membrane potential v_mem, clipped at every time step.
    shape: torch.Size
        Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
    record_states: bool
        When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.
    """

    def __init__(
        self,
        rec_connect: torch.nn.Module,
        spike_threshold: float = 1.0,
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        tau_syn: Optional[float] = None,
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        record_states: bool = False,
    ):
        super().__init__(
            rec_connect=rec_connect,
            tau_mem=np.inf,
            tau_syn=tau_syn,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            norm_input=False,
            record_states=record_states,
        )
        # deactivate tau_mem being learned
        self.tau_mem.requires_grad = False

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.pop("tau_mem")
        param_dict.pop("train_alphas")
        param_dict.pop("norm_input")
        return param_dict


class IAFSqueeze(IAF, SqueezeMixin):
    """
    IAF layer with 4-dimensional input (Batch*Time, Channel, Height, Width).

    Same as parent IAF class, only takes in squeezed 4D input (Batch*Time, Channel, Height, Width)
    instead of 5D input (Batch, Time, Channel, Height, Width) in order to be compatible with
    layers that can only take a 4D input, such as convolutional and pooling layers.
    """

    def __init__(
        self,
        batch_size=None,
        num_timesteps=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.squeeze_init(batch_size, num_timesteps)

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        return self.squeeze_forward(input_data, super().forward)

    @property
    def _param_dict(self) -> dict:
        return self.squeeze_param_dict(super()._param_dict)



from .quantize import Quantize, StochasticRounding
from .reset_mechanism import MembraneReset, MembraneSubtract
from .spike_generation import SingleSpike, MultiSpike, MaxSpike
from .surrogate_gradient_fn import (
    Heaviside,
    MultiGaussian,
    SingleExponential,
    PeriodicExponential,
)



import torch
import torch.nn as nn
from sinabs.layers import IAF, IAFSqueeze, IAFRecurrent
import sinabs.activation as sa
import pytest
import numpy as np


def test_iaf_basic():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = IAF()
    spike_output = layer(input_current)

    assert layer.does_spike
    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0
    assert "tau_mem" not in dict(layer.named_parameters()).keys()


def test_iaf_v_mem_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = IAF(record_states=True)
    spike_output = layer(input_current)

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert not layer.recordings["v_mem"].requires_grad
    assert "i_syn" not in layer.recordings.keys()


def test_iaf_i_syn_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = IAF(tau_syn=10.0, record_states=True)
    spike_output = layer(input_current)

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape
    assert not layer.recordings["v_mem"].requires_grad
    assert not layer.recordings["i_syn"].requires_grad


def test_iaf_recurrent_v_mem_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = IAFRecurrent(rec_connect=rec_connect, record_states=True)
    spike_output = layer(input_current)

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert not layer.recordings["v_mem"].requires_grad
    assert "i_syn" not in layer.recordings.keys()


def test_iaf_recurrent_i_syn_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = IAFRecurrent(tau_syn=10.0, rec_connect=rec_connect, record_states=True)
    spike_output = layer(input_current)

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape
    assert not layer.recordings["v_mem"].requires_grad
    assert not layer.recordings["i_syn"].requires_grad


def test_iaf_single_spike():
    batch_size, time_steps = 10, 100
    layer = IAF(spike_fn=sa.SingleSpike)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) * 100
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0
    assert torch.max(spike_output) == 1


def test_iaf_max_spike():
    batch_size, time_steps = 10, 100
    max_spikes = 3
    layer = IAF(spike_fn=sa.MaxSpike(max_spikes))
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) * 100
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0
    assert torch.max(spike_output) == max_spikes


def test_iaf_squeezed():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size * time_steps, 2, 7, 7)
    layer = IAFSqueeze(batch_size=batch_size)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_iaf_recurrent():
    batch_size, time_steps = 10, 100
    input_dimensions = (batch_size, time_steps, 2, 10)
    n_neurons = np.product(input_dimensions[2:])
    input_current = torch.ones(*input_dimensions) * 0.5

    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    rec_connect[1].weight = nn.Parameter(
        torch.ones(n_neurons, n_neurons) / n_neurons * 0.5
    )
    layer = IAFRecurrent(rec_connect=rec_connect)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_iaf_firing_rate():
    batch_size, time_steps, n_neurons = 5, 10, 5
    input_current = torch.zeros((batch_size, time_steps, n_neurons))
    input_current[:, 0] = 1

    layer = IAF()
    spikes = layer(input_current)

    assert layer.firing_rate == spikes.sum() / (batch_size * time_steps * n_neurons)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_iaf_on_gpu():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = IAF()

    layer = layer.to("cuda")
    layer(input_current.to("cuda"))


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_iaf_recurrent_on_gpu():
    batch_size, time_steps = 10, 100
    input_dimensions = (batch_size, time_steps, 2, 10)
    n_neurons = np.product(input_dimensions[2:])
    input_current = torch.ones(*input_dimensions)

    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = IAFRecurrent(rec_connect=rec_connect)

    layer = layer.to("cuda")
    layer(input_current.to("cuda"))



import torch
import sinabs.layers as sl


def test_flatten_time():
    data = torch.rand((5, 100, 2, 10, 10))
    flatten = sl.FlattenTime()
    output = flatten(data)

    assert output.shape == (500, 2, 10, 10)


def test_unflatten_time():
    data = torch.rand((500, 2, 10, 10))
    unflatten = sl.UnflattenTime(batch_size=5)
    output = unflatten(data)

    assert output.shape == (5, 100, 2, 10, 10)



# import torch
# import torch.nn as nn
# from pathlib import Path
# from sinabs.activation import ActivationFunction, MembraneSubtract

# MODELS_FOLDER = Path(__file__).resolve().parent / "models"


# def build_model():
#     from sinabs.from_torch import from_model

#     layers = [
#         nn.Conv2d(2, 8, kernel_size=2),
#         nn.AvgPool2d(2, stride=2),
#         nn.ReLU(),
#         nn.Conv2d(8, 10, kernel_size=3),
#         nn.AvgPool2d(3),
#         nn.ReLU(),
#     ]

#     ann = nn.Sequential(*layers)
#     net = from_model(ann)
#     return net


# def test_sinabs_model_to_onnx():
#     import onnx
#     from sinabs.onnx import print_onnx_model, get_graph, print_graph

#     net = build_model()
#     dummy_input = torch.zeros([1, 2, 64, 64])  # One time step
#     net(dummy_input)  # first pass to create all state variables
#     fname = MODELS_FOLDER / "snn.onnx"

#     torch.onnx.export(
#         net.spiking_model,
#         dummy_input,
#         fname,
#         input_names=["tchw_input"],
#         output_names=["tchw_output"],
#         dynamic_axes={"tchw_input": [0], "tchw_output": [0]},
#         verbose=False,
#     )
#     snn_model = onnx.load(fname)
#     ann_graph = get_graph(net, dummy_input)
#     print("................")
#     print_graph(ann_graph)
#     print("................")
#     print_onnx_model(snn_model)


# def test_graph_generation_ann():
#     import torchvision.models
#     import onnx
#     from sinabs.onnx import print_onnx_model

#     model = torchvision.models.resnet18()
#     dummy_input = torch.zeros([1, 3, 224, 224])
#     fname = MODELS_FOLDER / "resnet18.onnx"

#     torch.onnx.export(model, dummy_input, fname)
#     onnx_model = onnx.load(fname)

#     print_onnx_model(onnx_model)


# def test_onnx_sinabs_SpikingLayer():
#     import numpy as np
#     from sinabs.from_torch import from_model

#     layers = [
#         nn.ReLU(),
#     ]

#     ann = nn.Sequential(*layers)
#     net = from_model(ann)
#     inp = (torch.rand((50, 2, 32, 32)) > 0.5).float()
#     dummy = torch.zeros_like(inp)

#     net.spiking_model(dummy)

#     fname = MODELS_FOLDER / "test_spk.onnx"
#     torch.onnx.export(
#         net.spiking_model,
#         (dummy,),
#         fname,
#         export_params=True,
#         input_names=["tchw_input"],
#         output_names=["tchw_output"],
#         dynamic_axes={"tchw_input": [0], "tchw_output": [0]},
#         verbose=False,
#     )

#     # Run torch model
#     out_torch = net.spiking_model(inp)

#     # Run onnx model
#     import onnxruntime

#     session = onnxruntime.InferenceSession(str(fname))
#     ort_inputs = {session.get_inputs()[0].name: inp.numpy()}
#     ort_outs = session.run(None, ort_inputs)

#     np.testing.assert_allclose(
#         out_torch.detach().numpy(), ort_outs[0], rtol=1e-3, atol=1e-5
#     )


# def test_onnx_vs_sinabs_equivalence():
#     import numpy as np
#     net = build_model()

#     inp = (torch.rand((50, 2, 32, 32)) > 0.5).float()
#     dummy = torch.zeros_like(inp)

#     net.spiking_model(dummy)

#     fname = MODELS_FOLDER / "test_spk_net.onnx"
#     torch.onnx.export(
#         net.spiking_model,
#         (dummy,),
#         fname,
#         export_params=True,
#         input_names=["tchw_input"],
#         output_names=["tchw_output"],
#         dynamic_axes={"tchw_input": [0], "tchw_output": [0]},
#         verbose=False,
#     )

#     out_torch = net.spiking_model(inp)

#     import onnxruntime
#     session = onnxruntime.InferenceSession(str(fname))
#     ort_inputs = {session.get_inputs()[0].name: inp.numpy()}
#     ort_outs = session.run(None, ort_inputs)

#     np.testing.assert_allclose(
#         out_torch.detach().numpy(), ort_outs[0], atol=1e-5
#     )


# def test_threshold_subtract_onnx_eq():

#     class ThresholdTest(torch.nn.Module):
#         def __init__(self):
#             super().__init__()
#             self.activation_fn = ActivationFunction(spike_threshold=1.,
#                                                     reset_fn=MembraneSubtract(),
#                                                    )

#         def forward(self, inp):
#             state = {'v_mem': inp}
#             return self.activation_fn(state)

#     inp = (torch.rand(10)-0.5)*3
#     model = ThresholdTest()

#     torch_out = model(inp)
#     print(inp, "\nTorch:", torch_out)
#     fname = MODELS_FOLDER / "threshold_subtract.onnx"

#     torch.onnx.export(model, (inp,), fname, export_params=True,
#                       input_names=["t_input"],
#                       output_names=["t_output"],
#                       dynamic_axes={"t_input": [0], "t_output": [0]},
#                       verbose=False,
#                       )

#     import onnxruntime

#     session = onnxruntime.InferenceSession(str(fname))

#     ort_inputs = {session.get_inputs()[0].name: inp.numpy()}
#     ort_outs = session.run(None, ort_inputs)
#     print("Onnx:", ort_outs[0])

#     np.testing.assert_allclose(torch_out.detach().numpy(), ort_outs[0])



import copy
from typing import Callable, Optional, Tuple, Union
from warnings import warn
import torch
from torch import nn
from sinabs.activation import MembraneSubtract, MultiSpike, SingleExponential

from sinabs import Network
import sinabs.layers as sl

_backends = {"sinabs": sl}

try:
    import sinabs.exodus.layers as el
except ModuleNotFoundError:
    pass
else:
    _backends["exodus"] = el


def from_model(
    model,
    input_shape=None,
    spike_threshold=1.0,
    spike_fn: Callable = MultiSpike,
    reset_fn: Callable = MembraneSubtract(),
    surrogate_grad_fn: Callable = SingleExponential(),
    min_v_mem=-1.0,
    bias_rescaling=1.0,
    num_timesteps=None,
    batch_size=1,
    synops=False,
    add_spiking_output=False,
    backend="sinabs",
    kwargs_backend=None,
):
    """
    Converts a Torch model and returns a Sinabs network object.
    The modules in the model are analyzed, and a copy is returned, with all
    ReLUs, LeakyReLUs and NeuromorphicReLUs turned into SpikingLayers.

    Parameters
    ----------
    model:
        Torch model
    input_shape:
        If provided, the layer dimensions are computed. \
        Otherwise they will be computed at the first forward pass.
    spike_threshold:
        The membrane potential threshold for spiking (same for all layers).
    spike_fn: Callable
        The spike dynamics to determine the number of spikes out
    reset_fn: Callable
        The reset mechanism of the neuron (like reset to zero, or subtract)
    surrogate_grad_fn: Callable
        The surrogate gradient method for the spiking dynamics
    min_v_mem:
        The lower bound of the potential in (same for all layers).
    bias_rescaling:
        Biases are divided by this value.
    num_timesteps:
        Number of timesteps per sample. If None, `batch_size` \
        must be provided to seperate batch and time dimensions.
    batch_size:
        Must be provided if `num_timesteps` is None and is \
        ignored otherwise.
    synops:
        If True (default: False), register hooks for counting synaptic \
        operations during forward passes.
    add_spiking_output:
        If True (default: False), add a spiking layer \
        to the end of a sequential model if not present.
    backend:
        String defining the simulation backend (currently sinabs or exodus)
    kwargs_backend:
        Dict with additional kwargs for the simulation backend
    """
    return SpkConverter(
        input_shape=input_shape,
        spike_threshold=spike_threshold,
        spike_fn=spike_fn,
        reset_fn=reset_fn,
        surrogate_grad_fn=surrogate_grad_fn,
        min_v_mem=min_v_mem,
        bias_rescaling=bias_rescaling,
        batch_size=batch_size,
        num_timesteps=num_timesteps,
        synops=synops,
        add_spiking_output=add_spiking_output,
        backend=backend,
        kwargs_backend=kwargs_backend,
    ).convert(model)


class SpkConverter(object):
    """
    Converts a Torch model and returns a Sinabs network object.
    The modules in the model are analyzed, and a copy is returned, with all
    ReLUs, LeakyReLUs and NeuromorphicReLUs turned into SpikingLayers.

    Parameters
    ----------

    input_shape:
        If provided, the layer dimensions are computed. \
        Otherwise they will computed at the first forward pass.
    spike_threshold:
        The membrane potential threshold for spiking layers (same for all layers).
    spike_fn: Callable
        The spike dynamics to determine the number of spikes out
    reset_fn: Callable
        The reset mechanism of the neuron (like reset to zero, or subtract)
    surrogate_grad_fn: Callable
        The surrogate gradient method for the spiking dynamics
    min_v_mem:
        The lower bound of the potential in \
        convolutional and linear layers (same for all layers).
    bias_rescaling:
        Biases are divided by this value.
    num_timesteps:
        Number of timesteps per sample. If None, `batch_size` \
        must be provided to seperate batch and time dimensions.
    batch_size:
        Must be provided if `num_timesteps` is None and is \
        ignored otherwise.
    synops:
        If True (default: False), register hooks for counting synaptic \
        operations during foward passes.
    add_spiking_output:
        If True (default: False), add a spiking \
        layer to the end of a sequential model if not present.
    backend:
        String defining the simulation backend (currently sinabs or exodus)
    kwargs_backend:
        Dict with additional kwargs for the simulation backend
    """

    def __init__(
        self,
        input_shape: Optional[Tuple] = None,
        spike_threshold=1.0,
        spike_fn: Callable = MultiSpike(),
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        min_v_mem: float = -1.0,
        bias_rescaling: float = 1.0,
        num_timesteps: Optional[int] = None,
        batch_size: int = 1,
        synops: bool = False,
        add_spiking_output: bool = False,
        backend: str = "bptt",
        kwargs_backend: dict = None,
    ):
        self.min_v_mem = min_v_mem
        self.spike_threshold = spike_threshold
        self.spike_fn = spike_fn
        self.reset_fn = reset_fn
        self.surrogate_grad_fn = surrogate_grad_fn
        self.bias_rescaling = bias_rescaling
        self.batch_size = batch_size
        self.num_timesteps = num_timesteps
        self.synops = synops
        self.input_shape = input_shape
        self.add_spiking_output = add_spiking_output
        self.backend = backend
        self.kwargs_backend = kwargs_backend or dict()

    def relu2spiking(self):
        try:
            backend_module = _backends[self.backend]
        except KeyError:
            raise ValueError(
                f"Backend '{self.backend}' is not available. Available backends: "
                ", ".join(_backends.keys())
            )

        return backend_module.IAFSqueeze(
            spike_threshold=self.spike_threshold,
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            min_v_mem=self.min_v_mem,
            batch_size=self.batch_size,
            num_timesteps=self.num_timesteps,
            **self.kwargs_backend,
        ).to(self.device)

    def convert(self, model: nn.Module) -> Network:
        """
        Converts the Torch model and returns a Sinabs network object.
        Parameters
        ----------
        model:
            A torch module.

        Returns
        -------
        network:
            The Sinabs network object created by conversion.
        """
        spk_model = copy.deepcopy(model)
        # device is taken as the device of the first element of the input state_dict
        try:
            self.device = next(model.parameters()).device
        except StopIteration:
            self.device = torch.device("cpu")

        if self.add_spiking_output:
            # Add spiking output to sequential model
            if isinstance(spk_model, nn.Sequential) and not isinstance(
                spk_model[-1], (nn.ReLU, sl.NeuromorphicReLU)
            ):
                spk_model.add_module("Spiking output", nn.ReLU())
            else:
                warn(
                    "Spiking output can only be added to sequential models that do not end in a ReLU. No layer has been added."
                )

        self.convert_module(spk_model)
        network = Network(
            model,
            spk_model,
            input_shape=self.input_shape,
            synops=self.synops,
            batch_size=self.batch_size,
            num_timesteps=self.num_timesteps,
        )

        return network

    def convert_module(self, module):
        submodules = list(module.named_children())

        # iterate over the children
        for name, subm in submodules:
            # if it's one of the layers we're looking for, substitute it
            if isinstance(subm, (nn.ReLU, sl.NeuromorphicReLU)):
                setattr(module, name, self.relu2spiking())

            elif self.bias_rescaling != 1.0 and isinstance(
                subm, (nn.Linear, nn.Conv2d)
            ):
                if subm.bias is not None:
                    subm.bias.data = (
                        subm.bias.data.clone().detach() / self.bias_rescaling
                    )

            # if in turn it has children, go iteratively inside
            elif len(list(subm.named_children())):
                self.convert_module(subm)

            # otherwise we have a base layer of the non-interesting ones
            else:
                pass  # yes this is useless but it's for clarity



"""
========================
Integrate and Fire (IAF)
========================
:class:`~sinabs.layers.IAF` layer.
"""

from utils import plot_evolution
import sinabs.layers as sl
import torch


const_current = torch.ones((1, 100, 1)) * 0.03
single_current = torch.zeros((1, 100, 1))
single_current[:, 0] = 0.1

iaf_neuron = sl.IAF(record_states=True)
plot_evolution(iaf_neuron, const_current)


iaf_neuron = sl.IAF(tau_syn=15.0, record_states=True)
plot_evolution(iaf_neuron, single_current)



from pbr.version import VersionInfo

__version__ = VersionInfo("sinabs").release_string()

from .network import Network
from .synopcounter import SynOpCounter, SNNSynOpCounter
from .from_torch import from_model as convert_torch_ann

from .utils import (
    get_activations,
    get_network_activations,
)



"""
=================================
Exponential Leaky Layer (ExpLeak)
=================================
:class:`~sinabs.layers.ExpLeak` layer.
"""

from utils import plot_evolution
import sinabs.layers as sl
import torch


const_current = torch.ones((1, 100, 1)) * 0.03


exp_leak_neuron = sl.ExpLeak(tau_mem=60.0, record_states=True)
plot_evolution(exp_leak_neuron, const_current)



import torch
from copy import deepcopy


def test_deepcopy_iaf():
    from sinabs.layers import IAF, IAFSqueeze

    input_current = torch.rand(10, 10, 10)
    kwargs = dict(min_v_mem=-0.4)

    layer = IAF(**kwargs)
    layer_squeeze_nts = IAFSqueeze(**kwargs, num_timesteps=10)
    layer_squeeze_batch = IAFSqueeze(**kwargs, batch_size=10)

    for layer_orig in (layer, layer_squeeze_batch, layer_squeeze_nts):
        layer_orig(input_current)

        layer_copy = deepcopy(layer_orig)

        for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
            assert (p0 == p1).all()
            assert p0 is not p1
        for b0, b1 in zip(layer_orig.buffers(), layer_copy.buffers()):
            assert (b0 == b1).all()
            assert b0 is not b1

        assert layer_copy.spike_threshold == layer_orig.spike_threshold
        assert layer_copy.min_v_mem == layer_orig.min_v_mem
        if hasattr(layer_orig, "batch_size"):
            assert layer_orig.batch_size == layer_copy.batch_size
        if hasattr(layer_orig, "num_timesteps"):
            assert layer_orig.num_timesteps == layer_copy.num_timesteps


def test_deepcopy_iaf_uninitialized():
    from sinabs.layers import IAF, IAFSqueeze

    kwargs = dict(min_v_mem=-0.4)

    layer = IAF(**kwargs)
    layer_squeeze_nts = IAFSqueeze(**kwargs, num_timesteps=10)
    layer_squeeze_batch = IAFSqueeze(**kwargs, batch_size=10)

    for layer_orig in (layer, layer_squeeze_batch, layer_squeeze_nts):
        layer_copy = deepcopy(layer_orig)

        for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
            assert (p0 == p1).all()
            assert p0 is not p1
        for b0, b1 in zip(layer_orig.buffers(), layer_copy.buffers()):
            assert (b0 == b1).all()
            assert b0 is not b1

        assert layer_copy.spike_threshold == layer_orig.spike_threshold
        assert layer_copy.min_v_mem == layer_orig.min_v_mem
        if hasattr(layer_orig, "batch_size"):
            assert layer_orig.batch_size == layer_copy.batch_size
        if hasattr(layer_orig, "num_timesteps"):
            assert layer_orig.num_timesteps == layer_copy.num_timesteps


def test_deepcopy_lif():
    from sinabs.layers import LIF, LIFSqueeze

    for train_alphas in (True, False):

        input_current = torch.rand(10, 10, 10)
        kwargs = dict(tau_mem=torch.tensor(30.0), tau_syn=torch.tensor(10.0))

        kwargs["train_alphas"] = train_alphas

        layer = LIF(**kwargs)
        layer_squeeze_batch = LIFSqueeze(**kwargs, batch_size=10)
        layer_squeeze_nts = LIFSqueeze(**kwargs, num_timesteps=10)
        #         layer_recurrent = LIFRecurrent(**kwargs, rec_connect=torch.nn.Linear(10,10))

        for layer_orig in (layer, layer_squeeze_batch, layer_squeeze_nts):
            layer_orig(input_current)

            layer_copy = deepcopy(layer_orig)

            for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
                assert (p0 == p1).all()
                assert p0 is not p1
            for b0, b1 in zip(layer_orig.buffers(), layer_copy.buffers()):
                assert (b0 == b1).all()
                assert b0 is not b1

            assert layer_copy.spike_threshold == layer_orig.spike_threshold
            if train_alphas:
                assert layer_copy.alpha_mem == layer_orig.alpha_mem
            else:
                assert layer_copy.tau_mem == layer_orig.tau_mem
            assert layer_copy.spike_fn == layer_orig.spike_fn
            assert layer_copy.min_v_mem == layer_orig.min_v_mem
            assert layer_copy.train_alphas == layer_orig.train_alphas
            if hasattr(layer_orig, "batch_size"):
                assert layer_orig.batch_size == layer_copy.batch_size
            if hasattr(layer_orig, "num_timesteps"):
                assert layer_orig.num_timesteps == layer_copy.num_timesteps
            if hasattr(layer_orig, "rec_connect"):
                assert (
                    layer_orig.rec_connect.in_features
                    == layer_copy.rec_connect.in_features
                )


def test_deepcopy_lif_uninitialized():
    from sinabs.layers import LIF, LIFSqueeze

    for train_alphas in (True, False):
        kwargs = dict(tau_mem=torch.tensor(30.0), tau_syn=torch.tensor(10.0))

        kwargs["train_alphas"] = train_alphas

        layer = LIF(**kwargs)
        layer_squeeze_batch = LIFSqueeze(**kwargs, batch_size=10)
        layer_squeeze_nts = LIFSqueeze(**kwargs, num_timesteps=10)
        #         layer_recurrent = LIFRecurrent(**kwargs, rec_connect=torch.nn.Linear(10,10))

        for layer_orig in (layer, layer_squeeze_batch, layer_squeeze_nts):

            layer_copy = deepcopy(layer_orig)

            for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
                assert (p0 == p1).all()
                assert p0 is not p1
            for b0, b1 in zip(layer_orig.buffers(), layer_copy.buffers()):
                assert (b0 == b1).all()
                assert b0 is not b1

            assert layer_copy.spike_threshold == layer_orig.spike_threshold
            if train_alphas:
                assert layer_copy.alpha_mem == layer_orig.alpha_mem
            else:
                assert layer_copy.tau_mem == layer_orig.tau_mem
            assert layer_copy.spike_fn == layer_orig.spike_fn
            assert layer_copy.min_v_mem == layer_orig.min_v_mem
            assert layer_copy.train_alphas == layer_orig.train_alphas
            if hasattr(layer_orig, "batch_size"):
                assert layer_orig.batch_size == layer_copy.batch_size
            if hasattr(layer_orig, "num_timesteps"):
                assert layer_orig.num_timesteps == layer_copy.num_timesteps
            if hasattr(layer_orig, "rec_connect"):
                assert (
                    layer_orig.rec_connect.in_features
                    == layer_copy.rec_connect.in_features
                )


def test_deepcopy_alif():
    from sinabs.layers import ALIF

    kwargs = dict(
        tau_mem=20.0,
        tau_adapt=10.0,
        adapt_scale=1.3,
        min_v_mem=-0.4,
    )

    layer_orig = ALIF(**kwargs)

    layer_orig(torch.rand(10, 10, 10))

    layer_copy = deepcopy(layer_orig)

    for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
        assert (p0 == p1).all()
        assert p0 is not p1
    for (name, b0), (name, b1) in zip(
        layer_orig.named_buffers(), layer_copy.named_buffers()
    ):
        assert (b0 == b1).all()
        assert b0 is not b1

    assert layer_copy.min_v_mem == layer_orig.min_v_mem
    assert layer_copy.spike_fn == layer_orig.spike_fn
    assert layer_copy.tau_mem == layer_orig.tau_mem
    assert layer_copy.tau_adapt == layer_orig.tau_adapt
    assert layer_copy.adapt_scale == layer_orig.adapt_scale
    if hasattr(layer_orig, "batch_size"):
        assert layer_orig.batch_size == layer_copy.batch_size
    if hasattr(layer_orig, "num_timesteps"):
        assert layer_orig.num_timesteps == layer_copy.num_timesteps


def test_deepcopy_expleak():
    from sinabs.layers import ExpLeak, ExpLeakSqueeze

    layer = ExpLeak(tau_mem=10.0)
    layer_squeeze_nts = ExpLeakSqueeze(tau_mem=10.0, num_timesteps=10)
    layer_squeeze_batch = ExpLeakSqueeze(tau_mem=10.0, batch_size=10)

    for layer_orig in (layer, layer_squeeze_batch, layer_squeeze_nts):
        layer_orig(torch.rand(10, 10, 10))

        layer_copy = deepcopy(layer_orig)

        for p0, p1 in zip(layer_orig.parameters(), layer_copy.parameters()):
            assert (p0 == p1).all()
            assert p0 is not p1
        for b0, b1 in zip(layer_orig.buffers(), layer_copy.buffers()):
            assert (b0 == b1).all()
            assert b0 is not b1

        assert layer_copy.tau_mem == layer_orig.tau_mem
        if hasattr(layer_orig, "batch_size"):
            assert layer_orig.batch_size == layer_copy.batch_size
        if hasattr(layer_orig, "num_timesteps"):
            assert layer_orig.num_timesteps == layer_copy.num_timesteps



import torch
from typing import Callable


def alif_forward_single(
    input_data: torch.Tensor,
    alpha_mem: torch.Tensor,
    alpha_adapt: torch.Tensor,
    alpha_syn: torch.Tensor,
    adapt_scale: float,
    state: dict,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: float,
    b0: float,
    norm_input: bool,
):
    # if t_syn was provided, we're going to use synaptic current dynamics
    if alpha_syn is not None:
        state["i_syn"] = alpha_syn * (state["i_syn"] + input_data)
    else:
        state["i_syn"] = input_data

    if norm_input:
        state["i_syn"] = (1 - alpha_mem) * state["i_syn"]

    # Decay the membrane potential and add the input currents which are normalised by tau
    state["v_mem"] = alpha_mem * state["v_mem"] + state["i_syn"]

    # generate spikes and adjust v_mem
    input_tensors = [state[name] for name in spike_fn.required_states]
    state["spike_threshold"] = b0 + adapt_scale * state["b"]
    spikes = spike_fn.apply(*input_tensors, state["spike_threshold"], surrogate_grad_fn)

    state = reset_fn(spikes, state, state["spike_threshold"])

    # Decay the spike threshold and add adaptation factor to it.
    state["b"] = alpha_adapt * state["b"] + (1 - alpha_adapt) * spikes
    state["spike_threshold"] = b0 + adapt_scale * state["b"]

    # Clip membrane potential that is too low
    if min_v_mem is not None:
        state["v_mem"] = (
            torch.nn.functional.relu(state["v_mem"] - min_v_mem) + min_v_mem
        )

    return spikes, state


def alif_forward(
    input_data: torch.Tensor,
    alpha_mem: torch.Tensor,
    alpha_adapt: torch.Tensor,
    alpha_syn: torch.Tensor,
    adapt_scale: float,
    state: dict,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: float,
    b0: float,
    norm_input: bool,
    record_states: bool = False,
):
    time_steps = input_data.shape[1]
    state_names = list(state.keys())

    output_spikes = []
    recordings = []
    for step in range(time_steps):
        spikes, state = alif_forward_single(
            input_data=input_data[:, step],
            alpha_mem=alpha_mem,
            alpha_adapt=alpha_adapt,
            alpha_syn=alpha_syn,
            adapt_scale=adapt_scale,
            state=state,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            b0=b0,
            norm_input=norm_input,
        )
        output_spikes.append(spikes)
        if record_states:
            recordings.append(state)

    record_dict = {}
    if record_states:
        for state_name in state_names:
            record_dict[state_name] = torch.stack(
                [item[state_name].detach() for item in recordings], 1
            )
    return torch.stack(output_spikes, 1), state, record_dict


def alif_recurrent(
    input_data: torch.Tensor,
    alpha_mem: torch.Tensor,
    alpha_adapt: torch.Tensor,
    alpha_syn: torch.Tensor,
    adapt_scale: float,
    state: dict,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: float,
    rec_connect: torch.nn.Module,
    b0: float,
    norm_input: bool,
    record_states: bool = False,
):
    batch_size, n_time_steps, *trailing_dim = input_data.shape
    state_names = list(state.keys())

    output_spikes = []
    recordings = []
    rec_out = torch.zeros((batch_size, *trailing_dim), device=input_data.device)
    for step in range(n_time_steps):
        total_input = input_data[:, step] + rec_out

        spikes, state = alif_forward_single(
            input_data=total_input,
            alpha_mem=alpha_mem,
            alpha_adapt=alpha_adapt,
            alpha_syn=alpha_syn,
            adapt_scale=adapt_scale,
            state=state,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            b0=b0,
            norm_input=norm_input,
        )
        output_spikes.append(spikes)
        if record_states:
            recordings.append(state)

        # compute recurrent output that will be added to the input at the next time step
        rec_out = rec_connect(spikes).reshape((batch_size, *trailing_dim))

    record_dict = {}
    if record_states:
        for state_name in state_names:
            record_dict[state_name] = torch.stack(
                [item[state_name].detach() for item in recordings], 1
            )
    return torch.stack(output_spikes, 1), state, record_dict



import torch
import torch.nn as nn
from typing import Callable, Optional


class FlattenTime(nn.Flatten):
    """
    Utility layer which always flattens first two dimensions. Meant
    to convert a tensor of dimensions (Batch, Time, Channels, Height, Width)
    into a tensor of (Batch*Time, Channels, Height, Width).
    """

    def __init__(self):
        super().__init__(start_dim=0, end_dim=1)


class UnflattenTime(nn.Module):
    """
    Utility layer which always unflattens (expands) the first dimension into two separate ones.
    Meant to convert a tensor of dimensions (Batch*Time, Channels, Height, Width)
    into a tensor of (Batch, Time, Channels, Height, Width).
    """

    def __init__(self, batch_size: int):
        super().__init__()
        self.batch_size = batch_size

    def forward(self, x):
        num_time_steps = x.shape[0] // self.batch_size
        return x.unflatten(0, (self.batch_size, num_time_steps))


class SqueezeMixin:
    """
    Utility mixin class that will wrap the __init__ and forward call of other classes.
    The wrapped __init__ will provide two additional parameters batch_size and num_timesteps
    and the wrapped forward will unpack and repack the first dimension into batch and time.
    """

    def squeeze_init(self, batch_size: Optional[int], num_timesteps: Optional[int]):
        if not batch_size and not num_timesteps:
            raise TypeError("You need to specify either batch_size or num_timesteps.")
        if not batch_size:
            batch_size = -1
        if not num_timesteps:
            num_timesteps = -1
        self.batch_size = int(batch_size)
        self.num_timesteps = int(num_timesteps)

    def squeeze_forward(self, input_data: torch.Tensor, forward_method: Callable):
        inflated_input = input_data.reshape(
            self.batch_size, self.num_timesteps, *input_data.shape[1:]
        )
        inflated_output = forward_method(inflated_input)
        return inflated_output.flatten(start_dim=0, end_dim=1)

    def squeeze_param_dict(self, param_dict: dict) -> dict:
        param_dict.update(
            batch_size=self.batch_size,
            num_timesteps=self.num_timesteps,
        )
        return param_dict



from typing import Union, Optional
import torch
from .lif import LIF
from .reshape import SqueezeMixin


class ExpLeak(LIF):
    """
    Leaky Integrator layer.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = \\alpha V_{mem}(t) + (1-\\alpha)\\sum z(t)

    where :math:`\\alpha =  e^{-1/tau_{mem}}` and :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters
    ----------
    tau_mem: float
        Membrane potential time constant.
    min_v_mem: float or None
        Lower bound for membrane potential v_mem, clipped at every time step.
    train_alphas: bool
        When True, the discrete decay factor exp(-1/tau) is used for training rather than tau itself.
    shape: torch.Size
        Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
    norm_input: bool
        When True, normalise input current by tau. This helps when training time constants.
    record_states: bool
        When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.
    """

    def __init__(
        self,
        tau_mem: Union[float, torch.Tensor],
        shape: Optional[torch.Size] = None,
        train_alphas: bool = False,
        min_v_mem: Optional[float] = None,
        norm_input: bool = False,
        record_states: bool = False,
    ):
        super().__init__(
            tau_mem=tau_mem,
            tau_syn=None,
            spike_threshold=None,
            train_alphas=train_alphas,
            min_v_mem=min_v_mem,
            shape=shape,
            spike_fn=None,
            reset_fn=None,
            surrogate_grad_fn=None,
            norm_input=norm_input,
            record_states=record_states,
        )

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.pop("tau_syn")
        param_dict.pop("spike_fn")
        param_dict.pop("reset_fn")
        param_dict.pop("surrogate_grad_fn")
        param_dict.pop("spike_threshold")
        return param_dict


class ExpLeakSqueeze(ExpLeak, SqueezeMixin):
    """
    ExpLeak layer with 4-dimensional input (Batch*Time, Channel, Height, Width).

    Same as parent ExpLeak class, only takes in squeezed 4D input (Batch*Time, Channel, Height, Width)
    instead of 5D input (Batch, Time, Channel, Height, Width) in order to be compatible with
    layers that can only take a 4D input, such as convolutional and pooling layers.
    """

    def __init__(self, batch_size=None, num_timesteps=None, **kwargs):
        super().__init__(**kwargs)
        self.squeeze_init(batch_size, num_timesteps)

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        return self.squeeze_forward(input_data, super().forward)

    @property
    def _param_dict(self) -> dict:
        return self.squeeze_param_dict(super()._param_dict)



def test_crop2d():
    from sinabs.layers import Cropping2dLayer
    import torch

    sinabs_layer = Cropping2dLayer(((2, 3), (3, 1)))
    inp = torch.zeros((2, 2, 10, 10))
    assert sinabs_layer(inp).shape == (2, 2, 5, 6)



from dataclasses import dataclass
import torch
from typing import Optional, Callable, Union, List


class BackwardClass:
    @staticmethod
    def backward(ctx, grad_output: torch.tensor):
        """"""
        (v_mem,) = ctx.saved_tensors
        grad = ctx.surrogate_grad_fn(v_mem, ctx.spike_threshold)
        grad_input = grad_output * grad
        return grad_input, None, None


class MultiSpike(BackwardClass, torch.autograd.Function):
    """
    Autograd function that returns membrane potential integer-divided by spike threshold.
    Do not instantiate this class when passing as spike_fn (see example).
    Can be combined with different surrogate gradient functions.

    Example:
        >>> layer = sinabs.layers.LIF(spike_fn=MultiSpike, ...)
    """

    required_states: List[str] = ["v_mem"]

    @staticmethod
    def forward(
        ctx,
        v_mem: torch.Tensor,
        spike_threshold: Union[float, torch.Tensor],
        surrogate_grad_fn: Callable,
    ):
        """"""
        ctx.save_for_backward(v_mem.clone())
        ctx.spike_threshold = spike_threshold
        ctx.surrogate_grad_fn = surrogate_grad_fn
        spikes = (v_mem > 0) * torch.div(
            v_mem, spike_threshold, rounding_mode="trunc"
        ).float()
        return spikes


class MaxSpikeInner(BackwardClass, torch.autograd.Function):
    """
    Autograd function that returns membrane potential divided by
    spike threshold for a maximum number of spikes per time step.
    Can be combined with different surrogate gradient functions.
    """

    required_states: List[str] = ["v_mem", "max_num_spikes_per_bin"]

    @staticmethod
    def forward(
        ctx,
        v_mem: torch.Tensor,
        max_num_spikes_per_bin: Optional[int],
        spike_threshold: Union[float, torch.Tensor],
        surrogate_grad_fn: Callable,
    ):
        """"""
        ctx.save_for_backward(v_mem.clone())
        ctx.spike_threshold = spike_threshold
        ctx.surrogate_grad_fn = surrogate_grad_fn
        spikes = (v_mem > 0) * torch.div(
            v_mem, spike_threshold, rounding_mode="trunc"
        ).float()
        if max_num_spikes_per_bin is not None:
            spikes = torch.clamp(spikes, max=max_num_spikes_per_bin)
        return spikes


@dataclass
class MaxSpike:
    """
    Wrapper for MaxSpikeInner autograd function. This class needs to
    be instantiated when used as spike_fn. Notice the difference in example
    to Single/MultiSpike.

    Example:
        >>> layer = sinabs.layers.LIF(spike_fn=MaxSpike(max_num_spikes_per_bin=10), ...)
    """

    max_num_spikes_per_bin: Optional[int] = None

    def apply(
        self,
        v_mem: torch.Tensor,
        spike_threshold: Union[float, torch.Tensor],
        surrogate_grad_fn: Callable,
    ):
        return MaxSpikeInner.apply(
            v_mem, self.max_num_spikes_per_bin, spike_threshold, surrogate_grad_fn
        )

    @property
    def required_states(self):
        return ["v_mem"]


class SingleSpike(BackwardClass, torch.autograd.Function):
    """
    Autograd function that returns membrane potential divided by
    spike threshold for a maximum of one spike per time step.
    Do not instantiate this class when passing as spike_fn (see example).
    Can be combined with different surrogate gradient functions.

    Example:
        >>> layer = sinabs.layers.LIF(spike_fn=SingleSpike, ...)
    """

    required_states: List[str] = ["v_mem"]

    @staticmethod
    def forward(
        ctx,
        v_mem: torch.Tensor,
        spike_threshold: Union[float, torch.Tensor],
        surrogate_grad_fn: Callable,
    ):
        """"""
        ctx.save_for_backward(v_mem.clone())
        ctx.spike_threshold = spike_threshold
        ctx.surrogate_grad_fn = surrogate_grad_fn
        spikes = (v_mem - spike_threshold >= 0).float()
        return spikes



import warnings

import torch
from sinabs.layers import NeuromorphicReLU
from numpy import product


def synops_hook(layer, inp, out):
    assert len(inp) == 1, "Multiple inputs not supported for synops hook"
    inp = inp[0]
    layer.tot_in = inp.sum()
    layer.tot_out = out.sum()
    layer.synops = layer.tot_in * layer.fanout
    layer.tw = inp.shape[0]


class SNNSynOpCounter:
    """
    Counter for the synaptic operations emitted by all SpikingLayers in a
    spiking model.
    Note that this is automatically instantiated by `from_torch` and by
    `Network` if they are passed `synops=True`.

    Usage:
        counter = SNNSynOpCounter(my_spiking_model)

        output = my_spiking_model(input)  # forward pass

        synops_table = counter.get_synops()

    Arguments:
        model: Spiking model.
        dt: the number of milliseconds corresponding to a time step in the \
        simulation (default 1.0).
    """

    def __init__(self, model, dt=1.0):
        self.model = model
        self.handles = []
        self.dt = dt

        for layer in model.modules():
            self._register_synops_hook(layer)

    def _register_synops_hook(self, layer):
        if isinstance(layer, torch.nn.Conv2d):
            layer.fanout = (
                layer.out_channels * product(layer.kernel_size) / product(layer.stride)
            )
        elif isinstance(layer, torch.nn.Linear):
            layer.fanout = layer.out_features

        else:
            return None

        handle = layer.register_forward_hook(synops_hook)
        self.handles.append(handle)

    def get_synops(self) -> dict:
        """
        Method to compute a table of synaptic operations for the latest forward pass.

        The returned dictionary can be parsed into a table using pandas like so,
        ```
            synops_map = counter.get_synops()
            SynOps_dataframe = pandas.DataFrame.from_dict(synops_map, "index")
            SynOps_dataframe.set_index("Layer", inplace=True)
        ```

        NOTE: this may not be accurate in presence of average pooling.

        Returns:
            SynOps_map: A dictionary containing layer IDs and \
            respectively, for the latest forward pass performed, their:
                number of input spikes,
                fanout,
                synaptic operations,
                number of timesteps,
                total duration of simulation,
                number of synaptic operations per second.
        """
        SynOps_map = {}
        scale_facts = []
        for i, lyr in enumerate(self.model.modules()):
            if isinstance(lyr, torch.nn.AvgPool2d):
                if lyr.kernel_size != lyr.stride:
                    warnings.warn(
                        f"In order for the Synops counter to work accurately the pooling "
                        f"layers kernel size should match their strides. At the moment at layer {i}, "
                        f"the kernel_size = {lyr.kernel_size}, the stride = {lyr.stride}."
                    )
                ks = lyr.kernel_size
                scale_factor = ks**2 if isinstance(ks, int) else ks[0] * ks[1]
                scale_facts.append(scale_factor)
            if hasattr(lyr, "synops"):
                scale_factor = 1
                while len(scale_facts) != 0:
                    scale_factor *= scale_facts.pop()
                SynOps_map[i] = {
                    "Layer": i,
                    "In": lyr.tot_in * scale_factor,
                    "Fanout_Prev": lyr.fanout,
                    "SynOps": lyr.synops * scale_factor,
                    "N. timesteps": lyr.tw,
                    "Time window (ms)": lyr.tw * self.dt,
                    "SynOps/s": (lyr.synops * scale_factor) / lyr.tw / self.dt * 1000,
                }

        return SynOps_map

    def get_total_synops(self, per_second=False) -> float:
        """
        Faster method for computing total synaptic operations without using Pandas.

        NOTE: this may not be accurate in presence of average pooling.

        Arguments:
            per_second (bool, default False): if True, gives synops per second \
        instead of total synops in the last forward pass.

        Returns:
            synops: the total synops in the network, based on the last forward pass.
        """
        synops = 0.0
        for i, lyr in enumerate(self.model.modules()):
            if hasattr(lyr, "synops"):
                if per_second:
                    layer_synops = lyr.synops / lyr.tw / self.dt * 1000
                else:
                    layer_synops = lyr.synops

                synops = synops + layer_synops
        return synops

    def get_total_power_use(self, j_per_synop=1e-11):
        """
        Method to quickly get the total power use of the network, estimated
        over the latest forward pass.

        Arguments:
            j_per_synop: Energy use per synaptic operation, in joules.\
            Default 1e-11 J.

        Returns: estimated power in mW.
        """
        tot_synops_per_s = self.get_total_synops(per_second=True)
        power_in_mW = tot_synops_per_s * j_per_synop * 1000
        return power_in_mW

    def __del__(self):
        for handle in self.handles:
            handle.remove()


class SynOpCounter:
    """
    Counter for the synaptic operations emitted by all Neuromorphic ReLUs in an
    analog CNN model.

    Usage:
        counter = SynOpCounter(MyTorchModel.modules(), sum_activations=True)

        output = MyTorchModule(input)  # forward pass

        synop_count = counter()

    :param modules: list of modules, e.g. MyTorchModel.modules()
    :param sum_activations: If True (default), returns a single number of synops, otherwise a list of layer synops.
    """

    def __init__(self, modules, sum_activations=True):
        self.modules = []
        for module in modules:
            if isinstance(module, NeuromorphicReLU) and module.fanout > 0:
                self.modules.append(module)

        if len(self.modules) == 0:
            raise ValueError("No NeuromorphicReLU found in module list.")

        self.sum_activations = sum_activations
        # self.modules[1:] = []

    def __call__(self):
        synops = []
        for module in self.modules:
            synops.append(module.activity)

        if self.sum_activations:
            synops = torch.stack(synops).sum()
        return synops
